import { ChatOpenAI } from '@langchain/openai';
import type { BaseChatModel } from '@langchain/core/language_models/chat_models';
import type { Embeddings } from '@langchain/core/embeddings';
import {
  ChatPromptTemplate,
  MessagesPlaceholder,
  PromptTemplate,
} from '@langchain/core/prompts';
import { BaseMessage } from '@langchain/core/messages';
import { StringOutputParser } from '@langchain/core/output_parsers';
import LineListOutputParser from '../outputParsers/listLineOutputParser';
import LineOutputParser from '../outputParsers/lineOutputParser';
import { getDocumentsFromLinks } from '../utils/documents';
import { Document } from 'langchain/document';
import { searchSearxng, SearxngSearchResult } from '../searxng';
import path from 'node:path';
import fs from 'node:fs';
import computeSimilarity from '../utils/computeSimilarity';
import formatChatHistoryAsString from '../utils/formatHistory';
import eventEmitter from 'events';
import { StreamEvent } from '@langchain/core/tracers/log_stream';
import { EventEmitter } from 'node:stream';

export interface MetaSearchAgentType {
  searchAndAnswer: (
    message: string,
    history: BaseMessage[],
    llm: BaseChatModel,
    embeddings: Embeddings,
    optimizationMode: 'speed' | 'balanced' | 'quality',
    fileIds: string[],
    systemInstructions: string,
  ) => Promise<eventEmitter>;
}

interface Config {
  searchWeb: boolean;
  rerank: boolean;
  summarizer: boolean;
  rerankThreshold: number;
  queryGeneratorPrompt: string;
  responsePrompt: string;
  activeEngines: string[];
}

type SearchInput = {
  chat_history: BaseMessage[];
  query: string;
};

class MetaSearchAgent implements MetaSearchAgentType {
  private config: Config;
  private strParser = new StringOutputParser();

  constructor(config: Config) {
    this.config = config;
  }

  private async searchSources(
    llm: BaseChatModel,
    input: SearchInput,
    emitter: EventEmitter,
  ) {
    (llm as unknown as ChatOpenAI).temperature = 0;

    const chatPrompt = PromptTemplate.fromTemplate(
      this.config.queryGeneratorPrompt,
    );

    const processedChatPrompt = await chatPrompt.invoke({
      chat_history: formatChatHistoryAsString(input.chat_history),
      query: input.query,
    });

    const llmRes = await llm.invoke(processedChatPrompt);
    const messageStr = await this.strParser.invoke(llmRes);

    const linksOutputParser = new LineListOutputParser({
      key: 'links',
    });

    const questionOutputParser = new LineOutputParser({
      key: 'question',
    });

    const links = await linksOutputParser.parse(messageStr);
    let question = this.config.summarizer
      ? await questionOutputParser.parse(messageStr)
      : messageStr;

    if (question === 'not_needed') {
      return { query: '', docs: [] };
    }

    if (links.length > 0) {
      if (question.length === 0) {
        question = 'summarize';
      }

      let docs: Document[] = [];

      const linkDocs = await getDocumentsFromLinks({ links });

      const docGroups: Document[] = [];

      linkDocs.map((doc) => {
        const URLDocExists = docGroups.find(
          (d) =>
            d.metadata.url === doc.metadata.url && d.metadata.totalDocs < 10,
        );

        if (!URLDocExists) {
          docGroups.push({
            ...doc,
            metadata: {
              ...doc.metadata,
              totalDocs: 1,
            },
          });
        }

        const docIndex = docGroups.findIndex(
          (d) =>
            d.metadata.url === doc.metadata.url && d.metadata.totalDocs < 10,
        );

        if (docIndex !== -1) {
          docGroups[docIndex].pageContent =
            docGroups[docIndex].pageContent + `\n\n` + doc.pageContent;
          docGroups[docIndex].metadata.totalDocs += 1;
        }
      });

      await Promise.all(
        docGroups.map(async (doc) => {
          const res = await llm.invoke(`
        You are a web search summarizer, tasked with summarizing a piece of text retrieved from a web search. Your job is to summarize the 
        text into a detailed, 2-4 paragraph explanation that captures the main ideas and provides a comprehensive answer to the query.
        If the query is \"summarize\", you should provide a detailed summary of the text. If the query is a specific question, you should answer it in the summary.
        
        - **Journalistic tone**: The summary should sound professional and journalistic, not too casual or vague.
        - **Thorough and detailed**: Ensure that every key point from the text is captured and that the summary directly answers the query.
        - **Not too lengthy, but detailed**: The summary should be informative but not excessively long. Focus on providing detailed information in a concise format.

        The text will be shared inside the \`text\` XML tag, and the query inside the \`query\` XML tag.

        <example>
        1. \`<text>
        Docker is a set of platform-as-a-service products that use OS-level virtualization to deliver software in packages called containers. 
        It was first released in 2013 and is developed by Docker, Inc. Docker is designed to make it easier to create, deploy, and run applications 
        by using containers.
        </text>

        <query>
        What is Docker and how does it work?
        </query>

        Response:
        Docker is a revolutionary platform-as-a-service product developed by Docker, Inc., that uses container technology to make application 
        deployment more efficient. It allows developers to package their software with all necessary dependencies, making it easier to run in 
        any environment. Released in 2013, Docker has transformed the way applications are built, deployed, and managed.
        \`
        2. \`<text>
        The theory of relativity, or simply relativity, encompasses two interrelated theories of Albert Einstein: special relativity and general
        relativity. However, the word "relativity" is sometimes used in reference to Galilean invariance. The term "theory of relativity" was based
        on the expression "relative theory" used by Max Planck in 1906. The theory of relativity usually encompasses two interrelated theories by
        Albert Einstein: special relativity and general relativity. Special relativity applies to all physical phenomena in the absence of gravity.
        General relativity explains the law of gravitation and its relation to other forces of nature. It applies to the cosmological and astrophysical
        realm, including astronomy.
        </text>

        <query>
        summarize
        </query>

        Response:
        The theory of relativity, developed by Albert Einstein, encompasses two main theories: special relativity and general relativity. Special
        relativity applies to all physical phenomena in the absence of gravity, while general relativity explains the law of gravitation and its
        relation to other forces of nature. The theory of relativity is based on the concept of "relative theory," as introduced by Max Planck in
        1906. It is a fundamental theory in physics that has revolutionized our understanding of the universe.
        \`
        </example>

        Everything below is the actual data you will be working with. Good luck!

        <query>
        ${question}
        </query>

        <text>
        ${doc.pageContent}
        </text>

        Make sure to answer the query in the summary.
      `);

          const document = new Document({
            pageContent: res.content as string,
            metadata: {
              title: doc.metadata.title,
              url: doc.metadata.url,
            },
          });

          docs.push(document);
        }),
      );

      return { query: question, docs: docs };
    } else {
      question = question.replace(/<think>.*?<\/think>/g, '');

      const res = await searchSearxng(question, {
        language: 'en',
        engines: this.config.activeEngines,
      });

      const documents = res.results.map(
        (result) =>
          new Document({
            pageContent:
              result.content ||
              (this.config.activeEngines.includes('youtube')
                ? result.title
                : '') /* Todo: Implement transcript grabbing using Youtubei (source: https://www.npmjs.com/package/youtubei) */,
            metadata: {
              title: result.title,
              url: result.url,
              ...(result.img_src && { img_src: result.img_src }),
            },
          }),
      );

      return { query: question, docs: documents };
    }
  }

  private async performDeepResearch(
    llm: BaseChatModel,
    input: SearchInput,
    emitter: EventEmitter,
  ) {
    (llm as unknown as ChatOpenAI).temperature = 0;

    const queryGenPrompt = PromptTemplate.fromTemplate(
      this.config.queryGeneratorPrompt,
    );

    const formattedChatPrompt = await queryGenPrompt.invoke({
      chat_history: formatChatHistoryAsString(input.chat_history),
      query: input.query,
    });

    let i = 0;
    let currentQuery = await this.strParser.invoke(
      await llm.invoke(formattedChatPrompt),
    );
    const originalQuery = currentQuery;
    const pastQueries: string[] = [];
    const results: SearxngSearchResult[] = [];

    while (i < 10) {
      const res = await searchSearxng(currentQuery, {
        language: 'en',
        engines: this.config.activeEngines,
      });

      results.push(...res.results);

      const reflectorPrompt = PromptTemplate.fromTemplate(`
          You are an LLM that is tasked with reflecting on the results of a search query.

          ## Goal
          You will be given question of the user, a list of search results collected from the web to answer that question along with past queries made to collect those results. You have to analyze the results based on user's question and do the following:
          
          1. Identify unexplored areas or areas with less detailed information in the results and generate a new query that focuses on those areas. The new queries should be more specific and a similar query shall not exist in past queries which will be provided to you. Make sure to include keywords that you're looking for because the new query will be used to search the web for information on that topic. Make sure the query contains only 1 question and is not too long to ensure it is Search Engine friendly.
          2. You'll have to generate a description explaining what you are doing for example "I am looking for more information about X" or "Understanding how X works" etc. The description should be short and concise.
          
          ## Output format

          You need to output in XML format and do not generate any other text. ake sure to not include any other text in the output or start a conversation in the output. The output should be in the following format:

          <query>(query)</query>
          <description>(description)</description>

          ## Example
          Say the user asked "What is Llama 4 by Meta?" and let search results contain information about Llama 4 being an LLM and very little information about its features. You can output:

          <query>Llama 4 features</query> // Generate queries that capture keywords for SEO and not making words like "How", "What", "Why" etc.
          <description>Looking for new features in Llama 4</description>

          or something like

          <query>How is Llama 4 better than its previous generation models</query>
          <description>Understanding the difference between Llama 4 and previous generation models.</description>
          
          ## BELOW IS THE ACTUAL DATA YOU WILL BE WORKING WITH. IT IS NOT A PART OF EXAMPLES. YOU'LL HAVE TO GENERATE YOUR ANSWER BASED ON THIS DATA.
          <user_question>\n{question}\n</user_question>
          <search_results>\n{search_results}\n</search_results>
          <past_queries>\n{past_queries}\n</past_queries>

          Response:
      `);

      const formattedReflectorPrompt = await reflectorPrompt.invoke({
        question: originalQuery,
        search_results: results
          .map(
            (result) => `<result>${result.title} - ${result.content}</result>`,
          )
          .join('\n'),
        past_queries: pastQueries.map((q) => `<query>${q}</query>`).join('\n'),
      });

      const feedback = await this.strParser.invoke(
        await llm.invoke(formattedReflectorPrompt),
      );

      console.log(`Feedback: ${feedback}`);

      const queryOutputParser = new LineOutputParser({
        key: 'query',
      });

      const descriptionOutputParser = new LineOutputParser({
        key: 'description',
      });

      currentQuery = await queryOutputParser.parse(feedback);
      const description = await descriptionOutputParser.parse(feedback);
      console.log(`Query: ${currentQuery}`);
      console.log(`Description: ${description}`);

      pastQueries.push(currentQuery);
      ++i;
    }

    const uniqueResults: SearxngSearchResult[] = [];

    results.forEach((res) => {
      const exists = uniqueResults.find((r) => r.url === res.url);

      if (!exists) {
        uniqueResults.push(res);
      } else {
        exists.content += `\n\n` + res.content;
      }
    });

    const documents = uniqueResults /* .slice(0, 50) */
      .map(
        (r) =>
          new Document({
            pageContent: r.content || '',
            metadata: {
              title: r.title,
              url: r.url,
              ...(r.img_src && { img_src: r.img_src }),
            },
          }),
      );

    return documents;
  }

  private async streamAnswer(
    llm: BaseChatModel,
    fileIds: string[],
    embeddings: Embeddings,
    optimizationMode: 'speed' | 'balanced' | 'quality',
    systemInstructions: string,
    input: SearchInput,
    emitter: EventEmitter,
  ) {
    const chatPrompt = ChatPromptTemplate.fromMessages([
      ['system', this.config.responsePrompt],
      new MessagesPlaceholder('chat_history'),
      ['user', '{query}'],
    ]);

    let context = '';

    if (optimizationMode === 'speed' || optimizationMode === 'balanced') {
      let docs: Document[] | null = null;
      let query = input.query;

      if (this.config.searchWeb) {
        const searchResults = await this.searchSources(llm, input, emitter);

        query = searchResults.query;
        docs = searchResults.docs;
      }

      const sortedDocs = await this.rerankDocs(
        query,
        docs ?? [],
        fileIds,
        embeddings,
        optimizationMode,
      );

      emitter.emit(
        'data',
        JSON.stringify({ type: 'sources', data: sortedDocs }),
      );

      context = this.processDocs(sortedDocs);
    } else if (optimizationMode === 'quality') {
      let docs: Document[] = [];

      docs = await this.performDeepResearch(llm, input, emitter);

      emitter.emit('data', JSON.stringify({ type: 'sources', data: docs }));

      context = this.processDocs(docs);
    }

    const formattedChatPrompt = await chatPrompt.invoke({
      query: input.query,
      chat_history: input.chat_history,
      date: new Date().toISOString(),
      context: context,
      systemInstructions: systemInstructions,
    });

    const llmRes = await llm.stream(formattedChatPrompt);

    for await (const data of llmRes) {
      const messageStr = await this.strParser.invoke(data);

      emitter.emit(
        'data',
        JSON.stringify({ type: 'response', data: messageStr }),
      );
    }

    emitter.emit('end');
  }

  private async rerankDocs(
    query: string,
    docs: Document[],
    fileIds: string[],
    embeddings: Embeddings,
    optimizationMode: 'speed' | 'balanced' | 'quality',
  ) {
    if (docs.length === 0 && fileIds.length === 0) {
      return docs;
    }

    const filesData = fileIds
      .map((file) => {
        const filePath = path.join(process.cwd(), 'uploads', file);

        const contentPath = filePath + '-extracted.json';
        const embeddingsPath = filePath + '-embeddings.json';

        const content = JSON.parse(fs.readFileSync(contentPath, 'utf8'));
        const embeddings = JSON.parse(fs.readFileSync(embeddingsPath, 'utf8'));

        const fileSimilaritySearchObject = content.contents.map(
          (c: string, i: number) => {
            return {
              fileName: content.title,
              content: c,
              embeddings: embeddings.embeddings[i],
            };
          },
        );

        return fileSimilaritySearchObject;
      })
      .flat();

    if (query.toLocaleLowerCase() === 'summarize') {
      return docs.slice(0, 15);
    }

    const docsWithContent = docs.filter(
      (doc) => doc.pageContent && doc.pageContent.length > 0,
    );

    if (optimizationMode === 'speed' || this.config.rerank === false) {
      if (filesData.length > 0) {
        const [queryEmbedding] = await Promise.all([
          embeddings.embedQuery(query),
        ]);

        const fileDocs = filesData.map((fileData) => {
          return new Document({
            pageContent: fileData.content,
            metadata: {
              title: fileData.fileName,
              url: `File`,
            },
          });
        });

        const similarity = filesData.map((fileData, i) => {
          const sim = computeSimilarity(queryEmbedding, fileData.embeddings);

          return {
            index: i,
            similarity: sim,
          };
        });

        let sortedDocs = similarity
          .filter(
            (sim) => sim.similarity > (this.config.rerankThreshold ?? 0.3),
          )
          .sort((a, b) => b.similarity - a.similarity)
          .slice(0, 15)
          .map((sim) => fileDocs[sim.index]);

        sortedDocs =
          docsWithContent.length > 0 ? sortedDocs.slice(0, 8) : sortedDocs;

        return [
          ...sortedDocs,
          ...docsWithContent.slice(0, 15 - sortedDocs.length),
        ];
      } else {
        return docsWithContent.slice(0, 15);
      }
    } else if (optimizationMode === 'balanced') {
      const [docEmbeddings, queryEmbedding] = await Promise.all([
        embeddings.embedDocuments(
          docsWithContent.map((doc) => doc.pageContent),
        ),
        embeddings.embedQuery(query),
      ]);

      docsWithContent.push(
        ...filesData.map((fileData) => {
          return new Document({
            pageContent: fileData.content,
            metadata: {
              title: fileData.fileName,
              url: `File`,
            },
          });
        }),
      );

      docEmbeddings.push(...filesData.map((fileData) => fileData.embeddings));

      const similarity = docEmbeddings.map((docEmbedding, i) => {
        const sim = computeSimilarity(queryEmbedding, docEmbedding);

        return {
          index: i,
          similarity: sim,
        };
      });

      const sortedDocs = similarity
        .filter((sim) => sim.similarity > (this.config.rerankThreshold ?? 0.3))
        .sort((a, b) => b.similarity - a.similarity)
        .slice(0, 15)
        .map((sim) => docsWithContent[sim.index]);

      return sortedDocs;
    }

    return [];
  }

  private processDocs(docs: Document[]) {
    return docs
      .map(
        (_, index) =>
          `${index + 1}. ${docs[index].metadata.title} ${
            docs[index].pageContent
          }`,
      )
      .join('\n');
  }

  async searchAndAnswer(
    message: string,
    history: BaseMessage[],
    llm: BaseChatModel,
    embeddings: Embeddings,
    optimizationMode: 'speed' | 'balanced' | 'quality',
    fileIds: string[],
    systemInstructions: string,
  ) {
    const emitter = new eventEmitter();

    this.streamAnswer(
      llm,
      fileIds,
      embeddings,
      optimizationMode,
      systemInstructions,
      {
        chat_history: history,
        query: message,
      },
      emitter,
    );

    return emitter;
  }
}

export default MetaSearchAgent;
